@@ -8,6 +8,14 @@ use nom::{
     Err as NomErr, InputLength, Offset, Parser, Slice,
 };
 
+/// No-op function that typechecks that its argument is a parser. Used to
+/// ensure there are no accidentally missing type bounds on the `ParserExt`
+/// methods
+#[inline(always)]
+fn must_be_a_parser<I, O, E, P: Parser<I, O, E>>(parser: P) -> P {
+    parser
+}
+
 /// Additional postfix parser combinators, as a complement to [`Parser`].
 /// Mostly these are postfix versions of the combinators in [`nom::combinator`]
 /// and [`nom::sequence`], with some additional combinators original to
@@ -49,7 +57,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     #[inline]
     #[must_use = "Parsers do nothing unless used"]
     fn by_ref(&mut self) -> RefParser<Self> {
-        RefParser { parser: self }
+        must_be_a_parser(RefParser { parser: self })
     }
 
     /// Create a parser that must consume all of the input, or else return an
@@ -82,7 +90,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
         I: InputLength,
         E: ParseError<I>,
     {
-        AllConsuming { parser: self }
+        must_be_a_parser(AllConsuming { parser: self })
     }
 
     /// Create a parser that transforms `Incomplete` into `Error`.
@@ -114,7 +122,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
         I: Clone,
         E: ParseError<I>,
     {
-        Complete { parser: self }
+        must_be_a_parser(Complete { parser: self })
     }
 
     /// Create a parser that transforms `Error` into `Failure`. This will
@@ -140,7 +148,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     #[inline]
     #[must_use = "Parsers do nothing unless used"]
     fn cut(self) -> Cut<Self> {
-        Cut { parser: self }
+        must_be_a_parser(Cut { parser: self })
     }
 
     /// Create a parser that applies a mapping function `func` to the output
@@ -175,11 +183,52 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
         E: FromExternalError<I, E2>,
         I: Clone,
     {
-        MapRes {
+        must_be_a_parser(MapRes {
             parser: self,
             func,
             phantom: PhantomData,
-        }
+        })
+    }
+
+    /// Create a parser that applies a mapping function `func` to the output
+    /// of the subparser. Any errors from `func` will be transformed into
+    /// parse failures via [`FromExternalError`]. This will
+    /// end the parse immediately, even if there are other branches that
+    /// could occur.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// # use nom::{Err, Parser};
+    /// # use nom::error::{Error, ErrorKind};
+    /// use nom::character::complete::alphanumeric1;
+    /// use nom_supreme::parser_ext::ParserExt;
+    ///
+    /// let mut parser = alphanumeric1.map_res_cut(|s: &str| s.parse());
+    ///
+    /// assert_eq!(parser.parse("10 abc"), Ok((" abc", 10)));
+    /// assert_eq!(
+    ///     parser.parse("<===>"),
+    ///     Err(Err::Error(Error{input: "<===>", code: ErrorKind::AlphaNumeric})),
+    /// );
+    /// assert_eq!(
+    ///     parser.parse("abc abc"),
+    ///     Err(Err::Failure(Error{input: "abc abc", code: ErrorKind::MapRes})),
+    /// );
+    /// ```
+    #[inline]
+    #[must_use = "Parsers do nothing unless used"]
+    fn map_res_cut<F, O2, E2>(self, func: F) -> MapResCut<Self, F, O, E2>
+    where
+        F: FnMut(O) -> Result<O2, E2>,
+        E: FromExternalError<I, E2>,
+        I: Clone,
+    {
+        must_be_a_parser(MapResCut {
+            parser: self,
+            func,
+            phantom: PhantomData,
+        })
     }
 
     /// Make this parser optional; if it fails to parse, instead it returns
@@ -212,7 +261,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         I: Clone,
     {
-        Optional { parser: self }
+        must_be_a_parser(Optional { parser: self })
     }
 
     /// Replace this parser's output with the entire input that was consumed
@@ -241,10 +290,10 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         I: Clone + Slice<RangeTo<usize>> + Offset,
     {
-        Recognize {
+        must_be_a_parser(Recognize {
             parser: self.with_recognized(),
             phantom: PhantomData,
-        }
+        })
     }
 
     /// Return the parsed value, but also return the entire input that was
@@ -273,7 +322,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         I: Clone + Slice<RangeTo<usize>> + Offset,
     {
-        WithRecognized { parser: self }
+        must_be_a_parser(WithRecognized { parser: self })
     }
 
     /// Replace this parser's output with a clone of `value` every time it
@@ -321,11 +370,11 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     #[inline]
     #[must_use = "Parsers do nothing unless used"]
     fn value<T: Clone>(self, value: T) -> Value<T, Self, O> {
-        Value {
+        must_be_a_parser(Value {
             parser: self,
             value,
             phantom: PhantomData,
-        }
+        })
     }
 
     /// Require the output of this parser to pass a verifier function, or
@@ -362,10 +411,10 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
         I: Clone,
         E: ParseError<I>,
     {
-        Verify {
+        must_be_a_parser(Verify {
             parser: self,
             verifier,
-        }
+        })
     }
 
     /// Add some context to the parser. This context will be added to any
@@ -420,10 +469,10 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
         E: ContextError<I>,
         I: Clone,
     {
-        Context {
+        must_be_a_parser(Context {
             context,
             parser: self,
-        }
+        })
     }
 
     /// Add a terminator parser. The terminator will run after this parser,
@@ -451,11 +500,11 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         F: Parser<I, O2, E>,
     {
-        Terminated {
+        must_be_a_parser(Terminated {
             parser: self,
             terminator,
             phantom: PhantomData,
-        }
+        })
     }
 
     /// Make this parser precede another one. The successor parser will run
@@ -488,7 +537,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         F: Parser<I, O2, E>,
     {
-        successor.preceded_by(self)
+        must_be_a_parser(successor.preceded_by(self))
     }
 
     /// Make this parser preceded by another one. The `prefix` will run first,
@@ -522,11 +571,11 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         F: Parser<I, O2, E>,
     {
-        Preceded {
+        must_be_a_parser(Preceded {
             parser: self,
             prefix,
             phantom: PhantomData,
-        }
+        })
     }
 
     /// Make this parser delimited, requiring a `delimiter` as both a prefix and
@@ -558,11 +607,11 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         D: Parser<I, O2, E>,
     {
-        Delimited {
+        must_be_a_parser(Delimited {
             parser: self,
             delimiter,
             phantom: PhantomData,
-        }
+        })
     }
 
     /// Make this parser peeking: it runs normally but consumes no input.
@@ -589,7 +638,7 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
     where
         I: Clone,
     {
-        Peek { parser: self }
+        must_be_a_parser(Peek { parser: self })
     }
 
     /// Make this parser a negative lookahead: it will succeed if the subparser
@@ -618,10 +667,10 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
         I: Clone,
         E: ParseError<I>,
     {
-        Not {
+        must_be_a_parser(Not {
             parser: self,
             phantom: PhantomData,
-        }
+        })
     }
 
     /// Create a parser that parses something via [`FromStr`], using this
@@ -682,10 +731,81 @@ pub trait ParserExt<I, O, E>: Parser<I, O, E> + Sized {
         T: FromStr,
         E: FromExternalError<&'a str, T::Err>,
     {
-        FromStrParser {
+        must_be_a_parser(FromStrParser {
             parser: self,
             phantom: PhantomData,
-        }
+        })
+    }
+
+    /// Create a parser that parses something via [`FromStr`], using this
+    /// parser as a recognizer for the string to pass to
+    /// [`from_str`][FromStr::from_str]. This parser transforms any errors
+    /// from [`FromStr`] into [`Err::Failure`][NomErr::Failure], which will
+    /// end the overall parse immediately, even if there are other branches
+    /// that could be tried.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// # use nom::{Err, Parser, IResult};
+    /// # use nom::error::{Error, ErrorKind};
+    /// use nom::character::complete::alphanumeric1;
+    /// use nom_supreme::parser_ext::ParserExt;
+    ///
+    /// let mut parser = alphanumeric1.parse_from_str_cut();
+    ///
+    /// assert_eq!(parser.parse("123 abc"), Ok((" abc", 123)));
+    /// assert_eq!(
+    ///     parser.parse("<===>"),
+    ///     Err(Err::Error(Error{input: "<===>", code: ErrorKind::AlphaNumeric})),
+    /// );
+    /// assert_eq!(
+    ///     parser.parse("abc"),
+    ///     Err(Err::Failure(Error{input: "abc", code: ErrorKind::MapRes})),
+    /// );
+    /// ```
+    ///
+    /// # Parse error example
+    ///
+    /// If the [`FromStr`] parser fails, the error is recoverable from via
+    /// [`FromExternalError`]. In general, though, it's better practice to
+    /// ensure your recognizer won't allow invalid strings to be forwarded to
+    /// the [`FromStr`] parser
+    ///
+    /// ```rust
+    /// use std::num::ParseIntError;
+    /// use cool_asserts::assert_matches;
+    /// # use nom::{Err, Parser, IResult};
+    /// # use nom::error::{ErrorKind};
+    /// use nom::character::complete::alphanumeric1;
+    /// use nom_supreme::parser_ext::ParserExt;
+    /// use nom_supreme::error::{ErrorTree, BaseErrorKind};
+    ///
+    /// let mut parser = alphanumeric1.parse_from_str_cut();
+    ///
+    /// assert_matches!(parser.parse("123 abc"), Ok((" abc", 123)));
+    /// assert_matches!(
+    ///     parser.parse("abc"),
+    ///     Err(Err::Failure(ErrorTree::Base{
+    ///         location: "abc",
+    ///         kind: BaseErrorKind::External(err),
+    ///     })) => {
+    ///         let _err: &ParseIntError = err.downcast_ref().unwrap();
+    ///     },
+    /// );
+    /// ```
+    #[inline]
+    #[must_use = "Parsers do nothing unless used"]
+    fn parse_from_str_cut<'a, T>(self) -> FromStrCutParser<Self, T>
+    where
+        Self: Parser<&'a str, &'a str, E>,
+        T: FromStr,
+        E: FromExternalError<&'a str, T::Err>,
+    {
+        must_be_a_parser(FromStrCutParser {
+            parser: self,
+            phantom: PhantomData,
+        })
     }
 }
 
@@ -857,8 +977,9 @@ where
 {
     #[inline]
     fn parse(&mut self, input: I) -> nom::IResult<I, T, E> {
-        let (input, _) = self.parser.parse(input)?;
-        Ok((input, self.value.clone()))
+        self.parser
+            .parse(input)
+            .map(|(input, _)| (input, self.value.clone()))
     }
 }
 
@@ -1012,6 +1133,35 @@ where
     }
 }
 
+/// Parser which runs a fallible mapping function on the output of the
+/// subparser. Any errors returned by the mapping function are transformed
+/// into a parse failure.
+///
+#[derive(Debug, Clone, Copy)]
+pub struct MapResCut<P, F, O, E2> {
+    parser: P,
+    func: F,
+    phantom: PhantomData<(O, E2)>,
+}
+
+impl<P, F, I, O, E, O2, E2> Parser<I, O2, E> for MapResCut<P, F, O, E2>
+where
+    P: Parser<I, O, E>,
+    F: FnMut(O) -> Result<O2, E2>,
+    E: FromExternalError<I, E2>,
+    I: Clone,
+{
+    #[inline]
+    fn parse(&mut self, input: I) -> nom::IResult<I, O2, E> {
+        let (tail, value) = self.parser.parse(input.clone())?;
+
+        (self.func)(value)
+            .map(move |value| (tail, value))
+            .map_err(move |err| {
+                NomErr::Failure(E::from_external_error(input, NomErrorKind::MapRes, err))
+            })
+    }
+}
 /// Parser which runs a subparser but doesn't consume any input
 #[derive(Debug, Clone, Copy)]
 pub struct Peek<P> {
@@ -1082,3 +1232,32 @@ where
         }
     }
 }
+
+/// Parser which parses something via [`FromStr`], using a subparser as a
+/// recognizer for the string to pass to [`from_str`][FromStr::from_str].
+/// Returns [`Err::Failure`][NomErr::Failure] if the [`FromStr`] parse fails.
+#[derive(Debug, Clone, Copy)]
+pub struct FromStrCutParser<P, T> {
+    parser: P,
+    phantom: PhantomData<T>,
+}
+
+impl<'a, T, E, P> Parser<&'a str, T, E> for FromStrCutParser<P, T>
+where
+    P: Parser<&'a str, &'a str, E>,
+    T: FromStr,
+    E: FromExternalError<&'a str, T::Err>,
+{
+    #[inline]
+    fn parse(&mut self, input: &'a str) -> nom::IResult<&'a str, T, E> {
+        let (tail, value_str) = self.parser.parse(input)?;
+        match value_str.parse() {
+            Ok(value) => Ok((tail, value)),
+            Err(parse_err) => Err(NomErr::Failure(E::from_external_error(
+                input,
+                NomErrorKind::MapRes,
+                parse_err,
+            ))),
+        }
+    }
+}
